import re
import quopri
import hashlib
import pendulum
import pendulum.parsing.exceptions
from email_master.attachments import EmailAttachmentList, EmailAttachment
import codecs
import warnings
from email_master.compat import base64_decode, to_unicode, base64_decode_to_bytes
import uuid
import six
from types import FunctionType, MethodType
from email.header import Header
from binascii import Error as BinASCIIError

if six.PY2:
    from types import StringTypes, DictType, ListType
    UnicodeType = unicode
elif six.PY3:
    unicode = to_unicode
    DictType = dict
    ListType = list
    StringTypes = str
    UnicodeType = str

# Fields: Type, Default
FIELD_TYPES = {
    "cc": (StringTypes, u""),
    "bcc": (StringTypes, u""),
    "raw_content": (StringTypes, u""),
    "raw_headers": (StringTypes, u""),
    "reply_to": (StringTypes, u""),
    "sender": (StringTypes, u""),
    "plaintext_body": (StringTypes, u""),
    "html_body": (StringTypes, u""),
    "rtf_body": (StringTypes, u""),
    "subject": (StringTypes, u""),
    "type": (StringTypes, u""),
    "recipients": (StringTypes, u""),
    "headers": (ListType, []),
    "attachments": (EmailAttachmentList, EmailAttachmentList()),
    "id": (StringTypes, u""),
    "date": (StringTypes, u"")
}


class EmailParser(object):
    def __init__(self, email_data, filename="unknown", ignore_errors=False, exclude_attachment_extensions=None):
        self.filename = filename
        self.email_data = email_data
        self.ignore_errors = ignore_errors

        def to_lower(s):
            if isinstance(s, str):
                return str.lower(s)
            elif isinstance(s, unicode):
                return unicode.lower(s)
            else:
                raise TypeError("Invalid type detected in attachment exclusions! {}".format(s))

        self.exclude_attachment_extensions = tuple(map(to_lower, exclude_attachment_extensions)) if exclude_attachment_extensions else []

    def get_cc(self):
        raise NotImplementedError

    def get_bcc(self):
        raise NotImplementedError

    def get_raw_content(self):
        raise NotImplementedError

    def get_raw_headers(self):
        raise NotImplementedError

    def get_reply_to(self):
        raise NotImplementedError

    def get_sender(self):
        raise NotImplementedError

    def get_plaintext_body(self):
        raise NotImplementedError

    def get_html_body(self, decode_html=True):
        raise NotImplementedError

    def get_rtf_body(self):
        raise NotImplementedError

    def get_subject(self):
        raise NotImplementedError

    def get_type(self):
        raise NotImplementedError

    def get_recipients(self):
        raise NotImplementedError

    def get_headers(self):
        raise NotImplementedError

    def get_attachments(self):
        raise NotImplementedError

    def get_id(self):
        raise NotImplementedError

    def get_date(self):
        raise NotImplementedError

    def __getattribute__(self, item):
        """Wrap 'get_' calls to ensure some basic type checking"""
        attrib = super(EmailParser, self).__getattribute__(item)
        if isinstance(attrib, (FunctionType, MethodType)) and item.startswith("get_"):

            def wrapped(*args, **kwargs):
                output = attrib(*args, **kwargs)
                expected_type, default_val = FIELD_TYPES[item.replace("get_", "")]
                if isinstance(output, expected_type):
                    return output
                elif not output:  # Output resolves to None
                    warnings.warn("Invalid output type from Parser.{}! Expected '{}', got '{}'".format(
                        item,
                        expected_type,
                        type(output)
                    ))
                    return default_val
                else:
                    raise TypeError("Invalid type returned from Parser.{}! Expected '{}', got '{}'".format(
                        item,
                        expected_type,
                        type(output)
                    ))

            return wrapped
        else:
            return attrib

    @staticmethod
    def _parse_regex(regex, *bodies):
        text = ""
        for body in bodies:
            text = text + body if body else text
        val = sorted(list(set(re.findall(regex, text))))
        return val

    def _get_authenticated_sender(self, headers):
        return ",".join(
            self._parse_regex(r'(?:Authenticated sender:) ([a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+)', headers))

    def _get_clean_emails(self, email_data):
        valid_email_regex = r"([a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+)"
        return ",".join(self._parse_regex(valid_email_regex, email_data))

    def _try_date_format(self, date_data):
        try:
            # (=\?([^?.]+)\?([B|Q])\?((?:[A-Za-z0-9+\/]{4})*(?:[A-Za-z0-9+\/]{2}==|[A-Za-z0-9+\/]{3}=)?)\?=)
            dt = pendulum.parse(date_data)
            return dt.to_iso8601_string()
        except (pendulum.parsing.exceptions.ParserError, ValueError):
            return date_data

    def parse(self):
        result = {
            "result": u"failure", "attachments_sha1": "", "attachments_md5": "",
            "attach_info": "", "headers": "",
            "recipients": "", "subject": "", "text_body": "", "html_body": "", "type": "", "attachments_sha256": ""
        }

        attachment_list = self.get_attachments()
        if self.exclude_attachment_extensions:  # Filter out attachments and their hashes
            attachment_list = attachment_list.filter_by_filename(self.exclude_attachment_extensions)

        result["cc"] = self.get_cc()
        result["bcc"] = self.get_bcc()
        result["raw_headers"] = self.get_raw_headers()
        result["raw_content"] = self.get_raw_content()

        result["headers"] = ''.join(h[0] + ": " + h[1] + "\n" for h in self.get_headers())

        sender = self.get_sender()
        result["sender"] = sender if self._get_clean_emails(sender) else self._get_authenticated_sender(
            result['headers'])
        result["valid_sender_email"] = self._get_clean_emails(sender) or self._get_authenticated_sender(
            result['headers'])

        reply_to = self.get_reply_to()
        result["reply_to"] = reply_to
        result["valid_reply_to_email"] = self._get_clean_emails(reply_to)

        result["recipients"] = self.get_recipients()
        result["valid_recipients_email"] = self._get_clean_emails(self.get_recipients())

        result["subject"] = self.get_subject()
        result["text_body"] = self.get_plaintext_body()
        result["html_body"] = self.get_html_body()

        result["rtf_body"] = self.get_rtf_body()
        result["type"] = self.get_type()

        result["date"] = self._try_date_format(self.get_date())

        result["id"] = self.get_id()

        # Convert key, values to unicode (cleaning)
        for k, v in result.items():
            result[k] = v if v else u""
            if not isinstance(result[k], UnicodeType):
                result[k] = result.pop(k).decode(errors="replace")

        result["result"] = u"success"
        result.update(attachment_list.to_swimlane_output())

        result["orig_filename"] = self.filename
        return result


class EmailUtil(object):

    @staticmethod
    def parse_email_set(iterable_emails, parser_cls, parser_options=None, pre_hook=None, post_hook=None):
        """
        Parse a set of emails, given an iterable-like object of emails to parse
        This is useful for parsing lots of emails at once, also can add pre and post hooks for parsing

        """
        iterable_emails = iter(iterable_emails)
        results = []
        if not parser_options:
            parser_options = {}

        for email in iterable_emails:
            parser_inst = parser_cls(email, **parser_options)
            try:
                if pre_hook:  # Hook function for changing the email if needed
                    email = pre_hook(email)
                results.append(parser_inst.parse())
                if post_hook:  # Hook function for modifiying the email object (ie set read on a server)
                    email = post_hook(email)
            except Exception as e:
                if parser_options.get("ignore_errors", False):
                    results.append({
                        "error": str(e)
                    })
                else:
                    raise
        return results


    @staticmethod
    def try_decode(text):
        """
        Tries to decode a mime encoded-word syntax string
        See https://dmorgan.info/posts/encoded-word-syntax/ for more info

        :param text:
        :return:
        """
        mime_regex = r"=\?([^?.]+)\?"  # ?=<charset>?
        mime_regex += r"([B|Q])\?("  # Base64 or Quoted printable
        mime_regex += r"(?:[A-Za-z0-9+\/]{4})*(?:[A-Za-z0-9+\/]{2}==|[A-Za-z0-9+\/]{3}=)?"  # Base64 regex
        mime_regex += r"|"  # or
        mime_regex += r".+"  # quoted-printable regex (maybe todo?)
        mime_regex += r")\?="  # ending, ?=

        if not text:
            return u"(None)"

        result = u""

        if isinstance(text, Header):  # Py2&Py3 Compatibility
            text = str(text)

        for item in re.split(r'[\n\s]+', text):
            item = item.strip()
            match = re.match(mime_regex, item)
            if match:
                charset, encoding, encoded_text = match.groups()
                charset = EmailUtil.validate_charset(charset)

                if encoding is 'B':
                    try:
                        byte_str = base64_decode_to_bytes(encoded_text)
                    except (TypeError, BinASCIIError):  # Py3 compat
                        if six.PY3:
                            return encoded_text  # No need to decode this string in python3

                        byte_str = encoded_text  # Error with base64, just default to the text
                elif encoding is 'Q':
                    byte_str = quopri.decodestring(encoded_text)
                else:
                    # Can't decode this string, invalid encoding type
                    return text
                result = result + byte_str.decode(charset, errors="ignore")
            else:
                result = result + u" " + unicode(item, "utf-8", "ignore")

        return result.strip() or text  # Return result if it's been populated, else original text

    @staticmethod
    def validate_charset(charset):
        """
        Validate a string charset to ensure it exists
        Python standard encodings: https://docs.python.org/2.4/lib/standard-encodings.html
        Microsoft code number things: https://docs.microsoft.com/en-us/previous-versions/windows/desktop/bb322881(v%3Dvs.85)
        """

        if charset is None:
            return "utf-8"  # No data, have to default to utf

        def regex_match(regex_str, fixed_charset):
            def regex_fix(value):
                try:
                    if re.match(regex_str, value):
                        return fixed_charset
                except:  # Ignore all exceptions to make sure the charset test is fail-proof
                    pass
                finally:
                    return "definitelynotacharset"

            return regex_fix

        def regexify_keywords(keywords):
            re_str = "[^\s]{0,2}"
            spacer = "([-_$]?|[^\s]{0,2})"  # Space keywords out with things like -, _, $
            re_str += spacer.join(keywords)
            re_str += "[^\s]"  # Make sure it doesn't end with a space
            return re_str

        charset_synonyms = [  # These names mean the same thing but python doesn't know it
            ("134", "gb2312"),
            ("unicode", "utf-8"),
            ("windows-936", "gbk"),
            ("238", "cp852"),
            ("windows-1251", "cp1251")
        ]

        keyword_synonyms = [  # List of keywords -> charset, autogenerated using regex
            (["iso", "2022", "jp", "ext"], "iso-2022-jp-ext"),
            (["iso", "2022", "jp"], "iso-2022-jp"),
            (["iso", "2022", "kr"], "iso-2022-kr"),
        ]

        regex_synonyms = [
            # regex_match("myregex", "mycharset")  # Add regex charset data here
        ]

        # Add the keywords to the regex synonyms
        regex_synonyms.extend([regex_match(regexify_keywords(kws[0]), kws[1]) for kws in keyword_synonyms])

        possible_charset_fixes = [ # Add other charset fix functions here
            lambda x: x or "definitelynotacharset"  # No fix
        ]
        # Fixes from charset_synonyms (adds them to the possible_charset_fixes list)
        possible_charset_fixes.extend([lambda x: synm[1] if x.lower().startswith(synm[0]) or
                                                            x.lower().endswith(synm[0]) else x for synm in
                                       charset_synonyms])

        possible_charset_fixes.extend(regex_synonyms)  # Add the regex and keywords

        # Any fix here will only be run if the above fail
        possible_charset_fixes.extend([
            lambda x: x.replace("-", ""),  # Replace 'cp-850' with 'cp850'
            lambda x: "gb2312" if x.startswith("gb23") else x,
            lambda x: "utf-8"  # Fallback
        ])

        for fix in possible_charset_fixes:
            try:
                content_charset = fix(charset)
                codecs.lookup(content_charset)
                return content_charset
            except LookupError:
                continue
        raise Exception("Unknown charset: {}".format(charset))

    @staticmethod
    def check_content_charset(parser):
        """
        Try to get the email charset from a given email parser instance
        :param parser: EmailParser instance
        :return: most likely charset of the email
        """

        regex = r'(?:charset=\"?)([\w-]*)(?:\"+)'
        html_body = parser.get_html_body()
        headers = parser.get_headers()
        if headers:
            try:  # Try to get content charset from headers
                for header in headers:
                    if header[0].lower() == "content-type":
                        return EmailUtil.validate_charset(re.findall(regex, " ".join(header))[0])
            except IndexError:
                pass
        if html_body:
            try:  # Try to get content charset from html_body
                return EmailUtil.validate_charset(re.findall(regex, html_body)[0])
            except IndexError:
                pass
        return "utf-8"  # Default to utf-8 if we can't find it



# Sourced from https://git.autistici.org/ai/pgp-mime-lib/-/blob/78a371b5afb35ee223838130a0eb2d199f0d55eb/pgp_mime_lib/rfc3156.py
# From github.com/leapcode/leap_mail
#
# -*- coding: utf-8 -*-
# rfc3156.py
# Copyright (C) 2013 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""
Implements RFC 3156: MIME Security with OpenPGP.
"""
from __future__ import absolute_import

from email.mime.application import MIMEApplication
from email.mime.multipart import MIMEMultipart
from email import errors
from email.generator import (
    Generator,
    fcre,
    NL,
    _make_boundary,
)

import sys
import base64


PY2 = sys.version_info[0] == 2


if not PY2:
    from io import StringIO
    string_types = (str,)
else:
    from cStringIO import StringIO
    string_types = (str, unicode)


def base64_encodestring(data):
    if PY2:
        return base64.encodestring(str(data))
        # return base64.encodestring(data)
    r = base64.encodebytes(str(data).encode())
    return r


#
# A generator that solves http://bugs.python.org/issue14983
#

class RFC3156CompliantGenerator(Generator):
    """
    An email generator that addresses Python's issue #14983 for multipart
    messages.

    This is just a copy of email.generator.Generator which fixes the following
    bug: http://bugs.python.org/issue14983
    """

    def _handle_multipart(self, msg):
        """
        A multipart handling implementation that addresses issue #14983.

        This is just a copy of the parent's method which fixes the following
        bug: http://bugs.python.org/issue14983 (see the line marked with
        "(***)").

        :param msg: The multipart message to be handled.
        :type msg: email.message.Message
        """
        # The trick here is to write out each part separately, merge them all
        # together, and then make sure that the boundary we've chosen isn't
        # present in the payload.
        msgtexts = []
        subparts = msg.get_payload()
        if subparts is None:
            subparts = []
        elif isinstance(subparts, string_types):
            # e.g. a non-strict parse of a message with no starting boundary.
            self._fp.write(subparts)
            return
        elif not isinstance(subparts, list):
            # Scalar payload
            subparts = [subparts]
        for part in subparts:
            s = StringIO()
            g = self.clone(s)
            g.flatten(part, unixfrom=False)
            msgtexts.append(s.getvalue())
        # BAW: What about boundaries that are wrapped in double-quotes?
        boundary = msg.get_boundary()
        if not boundary:
            # Create a boundary that doesn't appear in any of the
            # message texts.
            alltext = NL.join(msgtexts)
            boundary = _make_boundary(alltext)
            msg.set_boundary(boundary)
        # If there's a preamble, write it out, with a trailing CRLF
        if msg.preamble is not None:
            preamble = msg.preamble
            if self._mangle_from_:
                preamble = fcre.sub('>From ', msg.preamble)
            self._fp.write(preamble + '\n')
        # dash-boundary transport-padding CRLF
        self._fp.write('--' + boundary + '\n')
        # body-part
        if msgtexts:
            self._fp.write(msgtexts.pop(0))
        # *encapsulation
        # --> delimiter transport-padding
        # --> CRLF body-part
        for body_part in msgtexts:
            # delimiter transport-padding CRLF
            self._fp.write('\n--' + boundary + '\n')
            # body-part
            self._fp.write(body_part)
        # close-delimiter transport-padding
        self._fp.write('\n--' + boundary + '--' + '\n')  # (***) Solve #14983
        if msg.epilogue is not None:
            self._fp.write('\n')
            epilogue = msg.epilogue
            if self._mangle_from_:
                epilogue = fcre.sub('>From ', msg.epilogue)
            self._fp.write(epilogue)


#
# Base64 encoding: these are almost the same as python's email.encoder
# solution, but a bit modified.
#

def _bencode(data):
    """
    Encode C{data} in base64.

    :param data: The string to be encoded.
    :type data: bytes
    """
    # We can't quite use base64.encodestring() since it tacks on a "courtesy
    # newline".  Blech!
    if not data:
        return data
    value = base64_encodestring(data)
    return value[:-1]


def encode_base64(msg):
    """
    Encode a non-multipart message's payload in Base64 (in place).

    This method modifies the message contents in place and adds or replaces an
    appropriate Content-Transfer-Encoding header.

    :param msg: The non-multipart message to be encoded.
    :type msg: email.message.Message
    """
    encoding = msg.get('Content-Transfer-Encoding', None)
    if encoding is not None:
        encoding = encoding.lower()
    # XXX Python's email module can only decode quoted-printable, base64 and
    # uuencoded data, so we might have to implement other decoding schemes in
    # order to support RFC 3156 properly and correctly calculate signatures
    # for multipart attachments (eg. 7bit or 8bit encoded attachments). For
    # now, if content is already encoded as base64 or if it is encoded with
    # some unknown encoding, we just pass.
    if encoding in [None, 'quoted-printable', 'x-uuencode', 'uue', 'x-uue', '8bit']:
        orig = msg.get_payload(decode=True)
        encdata = _bencode(orig)
        msg.set_payload(encdata)
        # replace or set the Content-Transfer-Encoding header.
        try:
            msg.replace_header('Content-Transfer-Encoding', 'base64')
        except KeyError:
            msg['Content-Transfer-Encoding'] = 'base64'
    elif encoding not in ('7bit', 'base64'):
        raise Exception('Unknown content-transfer-encoding: %s' % encoding)


def encode_base64_rec(msg):
    """
    Encode (possibly multipart) messages in base64 (in place).

    This method modifies the message contents in place.

    :param msg: The non-multipart message to be encoded.
    :type msg: email.message.Message
    """
    if not msg.is_multipart():
        encode_base64(msg)
    else:
        for sub in msg.get_payload():
            encode_base64_rec(sub)


#
# RFC 1847: multipart/signed and multipart/encrypted
#

class MultipartSigned(MIMEMultipart):
    """
    Multipart/Signed MIME message according to RFC 1847.

    2.1. Definition of Multipart/Signed

      (1)  MIME type name: multipart
      (2)  MIME subtype name: signed
      (3)  Required parameters: boundary, protocol, and micalg
      (4)  Optional parameters: none
      (5)  Security considerations: Must be treated as opaque while in
           transit

    The multipart/signed content type contains exactly two body parts.
    The first body part is the body part over which the digital signature
    was created, including its MIME headers.  The second body part
    contains the control information necessary to verify the digital
    signature.  The first body part may contain any valid MIME content
    type, labeled accordingly.  The second body part is labeled according
    to the value of the protocol parameter.

    When the OpenPGP digital signature is generated:

    (1)   The data to be signed MUST first be converted to its content-
          type specific canonical form.  For text/plain, this means
          conversion to an appropriate character set and conversion of
          line endings to the canonical <CR><LF> sequence.

    (2)   An appropriate Content-Transfer-Encoding is then applied; see
          section 3.  In particular, line endings in the encoded data
          MUST use the canonical <CR><LF> sequence where appropriate
          (note that the canonical line ending may or may not be present
          on the last line of encoded data and MUST NOT be included in
          the signature if absent).

    (3)   MIME content headers are then added to the body, each ending
          with the canonical <CR><LF> sequence.

    (4)   As described in section 3 of this document, any trailing
          whitespace MUST then be removed from the signed material.

    (5)   As described in [2], the digital signature MUST be calculated
          over both the data to be signed and its set of content headers.

    (6)   The signature MUST be generated detached from the signed data
          so that the process does not alter the signed data in any way.
    """

    def __init__(self, protocol, micalg, boundary=None, _subparts=None):
        """
        Initialize the multipart/signed message.

        :param boundary: the multipart boundary string. By default it is
            calculated as needed.
        :type boundary: str
        :param _subparts: a sequence of initial subparts for the payload. It
            must be an iterable object, such as a list. You can always
            attach new subparts to the message by using the attach() method.
        :type _subparts: iterable
        """
        MIMEMultipart.__init__(
            self, _subtype='signed', boundary=boundary,
            _subparts=_subparts)
        self.set_param('protocol', protocol)
        self.set_param('micalg', micalg)

    def attach(self, payload):
        """
        Add the C{payload} to the current payload list.

        Also prevent from adding payloads with wrong Content-Type and from
        exceeding a maximum of 2 payloads.

        :param payload: The payload to be attached.
        :type payload: email.message.Message
        """
        # second payload's content type must be equal to the protocol
        # parameter given on object creation
        if len(self.get_payload()) == 1:
            if payload.get_content_type() != self.get_param('protocol'):
                raise errors.MultipartConversionError(
                    'Wrong content type %s.' % payload.get_content_type)
        # prevent from adding more payloads
        if len(self._payload) == 2:
            raise errors.MultipartConversionError(
                'Cannot have more than two subparts.')
        MIMEMultipart.attach(self, payload)


class MultipartEncrypted(MIMEMultipart):
    """
    Multipart/encrypted MIME message according to RFC 1847.

    2.2. Definition of Multipart/Encrypted

      (1)  MIME type name: multipart
      (2)  MIME subtype name: encrypted
      (3)  Required parameters: boundary, protocol
      (4)  Optional parameters: none
      (5)  Security considerations: none

    The multipart/encrypted content type contains exactly two body parts.
    The first body part contains the control information necessary to
    decrypt the data in the second body part and is labeled according to
    the value of the protocol parameter.  The second body part contains
    the data which was encrypted and is always labeled
    application/octet-stream.
    """

    def __init__(self, protocol, boundary=None, _subparts=None):
        """
        :param protocol: The encryption protocol to be added as a parameter to
            the Content-Type header.
        :type protocol: str
        :param boundary: the multipart boundary string. By default it is
            calculated as needed.
        :type boundary: str
        :param _subparts: a sequence of initial subparts for the payload. It
            must be an iterable object, such as a list. You can always
            attach new subparts to the message by using the attach() method.
        :type _subparts: iterable
        """
        MIMEMultipart.__init__(
            self, _subtype='encrypted', boundary=boundary,
            _subparts=_subparts)
        self.set_param('protocol', protocol)

    def attach(self, payload):
        """
        Add the C{payload} to the current payload list.

        Also prevent from adding payloads with wrong Content-Type and from
        exceeding a maximum of 2 payloads.

        :param payload: The payload to be attached.
        :type payload: email.message.Message
        """
        # first payload's content type must be equal to the protocol parameter
        # given on object creation
        if len(self._payload) == 0:
            if payload.get_content_type() != self.get_param('protocol'):
                raise errors.MultipartConversionError(
                    'Wrong content type.')
        # second payload is always application/octet-stream
        if len(self._payload) == 1:
            if payload.get_content_type() != 'application/octet-stream':
                raise errors.MultipartConversionError(
                    'Wrong content type %s.' % payload.get_content_type)
        # prevent from adding more payloads
        if len(self._payload) == 2:
            raise errors.MultipartConversionError(
                'Cannot have more than two subparts.')
        MIMEMultipart.attach(self, payload)


#
# RFC 3156: application/pgp-encrypted, application/pgp-signed and
# application-pgp-signature.
#

class PGPEncrypted(MIMEApplication):
    """
    Application/pgp-encrypted MIME media type according to RFC 3156.

      * MIME media type name: application
      * MIME subtype name: pgp-encrypted
      * Required parameters: none
      * Optional parameters: none
    """

    def __init__(self, version=1):
        data = "Version: %d" % version
        MIMEApplication.__init__(self, data, 'pgp-encrypted')


class PGPSignature(MIMEApplication):
    """
    Application/pgp-signature MIME media type according to RFC 3156.

      * MIME media type name: application
      * MIME subtype name: pgp-signature
      * Required parameters: none
      * Optional parameters: none
    """
    def __init__(self, _data, name='signature.asc'):
        MIMEApplication.__init__(self, _data, 'pgp-signature',
                                 _encoder=lambda x: x, name=name)
        self.add_header('Content-Description', 'OpenPGP Digital Signature')


class PGPKeys(MIMEApplication):
    """
    Application/pgp-keys MIME media type according to RFC 3156.

      * MIME media type name: application
      * MIME subtype name: pgp-keys
      * Required parameters: none
      * Optional parameters: none
    """

    def __init__(self, _data):
        MIMEApplication.__init__(self, _data, 'pgp-keys')


from email_master.eml import EMLParser
from email_master.msg import MSGParser
from email_master.util import EmailParser, EmailUtil
from email_master.exceptions import InvalidDataException
from email_master.parser import EmailMasterParser



from email_master.compat import base64_decode_to_bytes, base64_encode_bytes
from pgpy.constants import SymmetricKeyAlgorithm, HashAlgorithm
from pgpy.errors import PGPError
from email_master.parser import EmailMasterParser
from email.encoders import encode_7or8bit
from pgpy import PGPKey, PGPMessage, PGPSignature
import email
from email.mime.application import MIMEApplication
from email.utils import collapse_rfc2231_value
from email import message_from_string
import email_master.rfc3156 as rfc3156
from datetime import datetime
import copy
import tempfile
import re
import six

if six.PY3:
    from io import StringIO
else:
    from cStringIO import StringIO


class PGPConfig(object):
    SIGNED = "signed"
    SIGNED_AND_ENCRYPT = "signedencrypt"
    ENCRYPTED = "encrypt"
    NONE = "none"

    def __undefined_attr(self, attr_name):
        def attr(self):
            raise ValueError("Property '{}' not defined and cannot be accessed".format(attr_name))

        return property(attr)

    def __init__(self, pgp_action=None, pgp_private_b64="", pgp_public_b64="", pgp_password=""):
        self.pgp_action = pgp_action
        self.key_pw = pgp_password

        if pgp_action in (PGPConfig.SIGNED, PGPConfig.SIGNED_AND_ENCRYPT, PGPConfig.ENCRYPTED):
            if not pgp_private_b64 or not pgp_public_b64:
                raise ValueError("Must provice PGP private and public keys to sign/encrypt!")
        elif pgp_action == PGPConfig.NONE:
            pass
        elif pgp_action is not None:
            raise ValueError(
                "Invalid pgp_action '{}'! Must be None, PGPConfig.SIGNED PGPConfig.SIGNED_AND_ENCRYPT, or PGPConfig.ENCRYPT!".format(
                    pgp_action))

        self.priv_key, self.priv_keystore = self._key_from_b64(pgp_private_b64)
        self.pub_key, self.pub_keystore = self._key_from_b64(pgp_public_b64)

        if self.priv_key:
            try:
                with self.priv_key.unlock(self.key_pw):
                    pass
            except Exception as e:
                raise ValueError("Invalid privkey or password, key unlock failed! Error: '{}'".format(str(e)))
        else:
            setattr(PGPConfig, "priv_key", self.__undefined_attr("Private Key"))

        if not self.pub_key:
            setattr(PGPConfig, "pub_key", self.__undefined_attr("Public Key"))

    def _key_from_b64(self, data):
        if not data:
            return None, None

        f = tempfile.NamedTemporaryFile(suffix=b'')
        dd = base64_decode_to_bytes(data)
        f.write(dd)
        f.seek(0)
        key, keystore = PGPKey.from_file(f.name)
        f.close()
        return key, keystore

    def _find_pgp_sig(self, msg_obj):
        attachments = len(msg_obj.get_payload())
        for idx in range(attachments):
            try:
                return list(self.pub_key.verify(msg_obj.get_payload(0).as_string(), PGPSignature.from_blob(msg_obj.get_payload(idx).get_payload())).good_signatures)
            except PGPError:
                continue  # Invalid signature
            except TypeError:
                continue  # Invalid body type
        return []  # Couldn't find an attachment with a valid signature

    def unlock(self, msg_as_string):
        mime_msg = CustomMIMEWrapper(msg_as_string)
        is_encrypted = mime_msg.is_encrypted()
        is_verified = False
        fingerprints = []
        message_type = "Encrypted" if is_encrypted else "Cleartext"

        if is_encrypted:
            with self.priv_key.unlock(self.key_pw) as unlocked_key:
                decrypted = mime_msg.decrypt(unlocked_key)
                if not decrypted:
                    raise ValueError("No decrypted message returned from mime_msg.decrypt!")
                mime_msg = decrypted

        is_signed = mime_msg.is_signed()  # Wait until message is possibly decrypted
        if is_signed:

            message_type += " & Signed"
            sigs = self._find_pgp_sig(mime_msg.msg)
            is_verified = len(sigs) > 0
            if is_verified:
                message_type += " & Verified"
            fingerprints = [str(sig.by.fingerprint) for sig in sigs]

        email_data = EmailMasterParser(base64_encode_bytes(str(mime_msg.msg).encode())).parse()

        email_data.update({
            "message_type": message_type,
            "is_verified": is_verified,
            "is_signed": is_signed,
            "is_encrypted": is_encrypted,
            "fingerprints": fingerprints
        })
        return email_data

    def _format_signed_payload(self, mime_message, payload, signed_payload, dt):
        bound = "wECpsh8cn1DyqgthMDFfJ9avZxL6XeJ90"  # TODO generate this
        new_headers = list(filter(lambda x: x[0] != "Content-Type", email.message_from_string(payload)._headers))

        signed_payload = signed_payload.as_string().replace("\n", "\r\n")

        header_data = "\r\n".join(["{}: {}".format(b[0], b[1]) for b in new_headers])
        newmsg = header_data + \
                 "\r\nContent-Type: multipart/signed; micalg=pgp-sha256; " + \
                 "protocol=\"application/pgp-signature\";" + \
                 "boundary=\"{bound}\"\r\n--{bound}\r\n".format(bound=bound) + \
                 payload + \
                 "\r\n--{bound}\r\n".format(bound=bound) + \
                 signed_payload + \
                 "--{}--".format(bound)
        return newmsg

    def lock(self, mime_message):
        if self.pgp_action is None:
            return mime_message
        elif self.pgp_action == PGPConfig.SIGNED:
            with self.priv_key.unlock(self.key_pw) as k:
                dt = datetime.utcnow()
                payload, signed_payload = mime_message.sign(k, hash=HashAlgorithm.SHA256, created=dt)
                return self._format_signed_payload(mime_message, payload, signed_payload, dt)
        elif self.pgp_action == PGPConfig.SIGNED_AND_ENCRYPT:
            with self.priv_key.unlock(self.key_pw) as k:
                mime = mime_message.sign_encrypt(k, list(self.pub_keystore.values()))
                return str(mime.msg)
        elif self.pgp_action == PGPConfig.ENCRYPTED:
            with self.priv_key.unlock(self.key_pw) as k:
                mime = mime_message.encrypt(k, list(self.pub_keystore.values()))
                return str(mime.msg)


class CustomMIMEWrapper(object):
    """PGP/MIME (RFC1847 + RFC3156) compliant wrapper."""
    _signature_subtype = 'pgp-signature'
    _encryption_subtype = 'pgp-encrypted'
    _keys_subtype = 'pgp-keys'
    _signed_type = 'application/' + _signature_subtype
    _encrypted_type = 'application/' + _encryption_subtype
    _keys_type = 'application/' + _keys_subtype
    _signed_multipart = 'multipart/signed'
    _encrypted_multipart = 'multipart/encrypted'
    _signature_preamble = \
        'This is an OpenPGP/MIME signed message (RFC 4880 and 3156)'
    _encryption_preamble = \
        'This is an OpenPGP/MIME encrypted message (RFC 4880 and 3156)'

    def __init__(self, msg):
        self.msg = msg

    def get_payload(self):
        yield self.msg.as_string()

    def _is_mime(self):
        is_multipart = self.msg.is_multipart()
        payloads = len(self.msg.get_payload()) if self.msg.get_payload() else 0
        return is_multipart and payloads == 2

    def _micalg(self, hash_algo):
        algs = {
            HashAlgorithm.MD5: 'md5',
            HashAlgorithm.SHA1: 'sha1',
            HashAlgorithm.RIPEMD160: 'ripemd160',
            HashAlgorithm.SHA256: 'sha256',
            HashAlgorithm.SHA384: 'sha384',
            HashAlgorithm.SHA512: 'sha512',
            HashAlgorithm.SHA224: 'sha224'
        }
        return 'pgp-' + algs[hash_algo]

    def openpgp_mangle_for_signature(self, msg):
        """Return a message suitable for signing.
        Encodes multipart message parts in msg as base64, then renders the
        message to string enforcing the right newline conventions. The
        returned value is suitable for signing according to RFC 3156.
        The incoming message is modified in-place.
        """
        rfc3156.encode_base64_rec(msg)
        fp = StringIO()
        g = rfc3156.RFC3156CompliantGenerator(
            fp, mangle_from_=False, maxheaderlen=76)

        g.flatten(msg)

        s = re.sub('\r?\n', '\r\n', fp.getvalue())
        if msg.is_multipart():
            if not s.endswith('\r\n'):
                s += '\r\n'
        return s

    def _wrap_signed(self, msg, signature):
        self.msg.set_payload([])
        self.msg.attach(msg)
        self.msg.set_type(CustomMIMEWrapper._signed_multipart)
        self.msg.set_param('micalg', self._micalg(signature.hash_algorithm))
        self.msg.set_param('protocol', CustomMIMEWrapper._signed_type)
        self.msg.preamble = CustomMIMEWrapper._signature_preamble
        second_part = MIMEApplication(_data=str(signature),
                                      _subtype=CustomMIMEWrapper._signature_subtype,
                                      _encoder=encode_7or8bit,
                                      name='signature.asc')
        second_part.add_header('Content-Description', 'OpenPGP digital signature')
        second_part.add_header('Content-Disposition', 'attachment', filename='signature.asc')
        self.msg.attach(second_part)
        return second_part

    def get_encrypted(self):
        try:
            msg = PGPMessage.from_blob(self.msg.get_payload(1).get_payload())
        except:
            return
        yield msg

    @staticmethod
    def copy_headers(from_msg, to_msg, overwrite=False):
        for key, value in from_msg.items():
            if overwrite:
                del to_msg[key]
            if key not in to_msg:
                to_msg[key] = value
        if to_msg.get_unixfrom() is None or overwrite:
            to_msg.set_unixfrom(from_msg.get_unixfrom())
        if (hasattr(from_msg, 'original_size')
                and (getattr(to_msg, 'original_size', None) is None
                     or overwrite)):
            to_msg.original_size = from_msg.original_size

    def decrypt(self, key):
        pmsg = next(iter(self.get_encrypted()))
        decrypted = key.decrypt(pmsg)

        dmsg = decrypted.message
        if isinstance(dmsg, bytearray):
            dmsg = dmsg.decode(decrypted.charset or 'utf-8')

        out = message_from_string(dmsg)
        if decrypted.is_signed:
            signature = next(iter(decrypted.signatures))
            self._wrap_signed(out, signature)
        else:
            self.msg.set_payload(out.get_payload())
            self.copy_headers(out, self.msg, True)
        return self

    def sign(self, key, **kwargs):
        payload = self.openpgp_mangle_for_signature(self.msg)

        signature = key.sign(payload, **kwargs)
        original_msg = copy.deepcopy(self.msg)
        return payload, self._wrap_signed(original_msg, signature)

    def _wrap_encrypted(self, payload):
        self.msg.set_payload([])
        self.msg.set_type(CustomMIMEWrapper._encrypted_multipart)
        self.msg.set_param('protocol', CustomMIMEWrapper._encrypted_type)
        self.msg.preamble = CustomMIMEWrapper._encryption_preamble
        first_part = MIMEApplication(_data='Version: 1',
                                     _subtype=CustomMIMEWrapper._encryption_subtype,
                                     _encoder=encode_7or8bit)
        first_part.add_header('Content-Description',
                              'PGP/MIME version identification')
        self.msg.attach(first_part)
        second_part = MIMEApplication(_data=str(payload),
                                      _subtype='octet-stream',
                                      _encoder=encode_7or8bit,
                                      name='encrypted.asc')
        second_part.add_header('Content-Description',
                               'OpenPGP encrypted message')
        second_part.add_header('Content-Disposition', 'inline',
                               filename='encrypted.asc')
        self.msg.attach(second_part)

    def _encrypt(self, pmsg, keys, cipher, **kwargs):
        emsg = copy.copy(pmsg)
        if len(keys) == 1:
            emsg = keys[0].encrypt(emsg, cipher=cipher, **kwargs)
        else:
            session_key = cipher.gen_key()
            for key in keys:
                emsg = key.encrypt(emsg, cipher=cipher,
                                   sessionkey=session_key,
                                   **kwargs)
            del session_key
        return emsg

    def encrypt(self, key, keys, **kwargs):
        hash = kwargs.get("hash", None)
        if not kwargs.get("cipher"):
            kwargs["cipher"] = SymmetricKeyAlgorithm.AES256

        if len(keys) == 0:
            raise ValueError('At least one key necessary.')

        payload = next(iter(self.get_payload()))
        pmsg = PGPMessage.new(payload)
        pmsg = self._encrypt(pmsg, keys, **kwargs)
        self._wrap_encrypted(pmsg)
        return self

    def sign_encrypt(self, key, keys, **kwargs):
        hash = kwargs.get("hash", None)

        if not kwargs.get("cipher"):
            kwargs["cipher"] = SymmetricKeyAlgorithm.AES256

        if len(keys) == 0:
            raise ValueError('At least one key necessary.')

        payload = next(iter(self.get_payload()))
        pmsg = PGPMessage.new(payload)
        pmsg |= key.sign(pmsg, hash=hash)
        pmsg = self._encrypt(pmsg, keys, **kwargs)
        self._wrap_encrypted(pmsg)
        return self

    def is_signed(self):
        if not self._is_mime():
            return False

        return collapse_rfc2231_value(self.msg.get_payload(1).get_content_type()) == CustomMIMEWrapper._signed_type and \
               self.msg.get_content_subtype() == "signed"

    def is_encrypted(self):
        if not self._is_mime():
            return False
        first_part = self.msg.get_payload(0).as_string()
        first_type = self.msg.get_payload(0).get_content_type()
        second_type = self.msg.get_payload(1).get_content_type()
        content_subtype = self.msg.get_content_subtype()
        protocol_param = collapse_rfc2231_value(self.msg.get_param('protocol', ''))
        return ('Version: 1' in first_part and
                first_type == CustomMIMEWrapper._encrypted_type and
                second_type == 'application/octet-stream' and
                content_subtype == 'encrypted' and
                protocol_param == CustomMIMEWrapper._encrypted_type)



from email_master.messages import EmailMasterMessage
from email_master.pgp import CustomMIMEWrapper
from email_master.compat import base64_encode, to_unicode
from email_master.compat import MIMEMultipart, EmailMessage
import mimetypes
import smtplib, ssl
import six

if six.PY2:
    from email.MIMEText import MIMEText
elif six.PY3:
    StringTypes = str
    unicode = to_unicode


class EMLMessage(EmailMasterMessage):
    def __init__(self, pgp_config=None):
        if six.PY3:
            super(EMLMessage, self).__init__(EmailMessage(), pgp_config)
        else:
            super(EMLMessage, self).__init__(MIMEMultipart(), pgp_config)

    def test_conn(self, username, password, host, port, verify_conn=True):
        conn = smtplib.SMTP(host, port, timeout=5)
        conn.ehlo()
        if verify_conn:
            context = ssl.create_default_context()
        else:
            context = ssl._create_unverified_context()

        if six.PY3:
            conn.starttls(context=context)
        else:
            conn.starttls()
            conn.ehlo()

        conn.login(username, password)
        conn.noop()
        return conn

    def send(self, username, password, host, port, verify_conn=True):
        for attch in self._attachments.attachments:  # Add attachments before sending
            fn = attch.filename
            data = attch.raw_data

            ctype_and_encoding = mimetypes.guess_extension(fn)
            if ctype_and_encoding is not None and len(ctype_and_encoding) == 2:
                ctype, encoding = ctype_and_encoding
            else:
                ctype = None
                encoding = None

            if ctype is None or encoding is not None:
                ctype = "application/octet-stream"
            maintype, subtype = ctype.split("/", 1)
            if six.PY3:
                self.message_obj.add_attachment(
                    data,
                    maintype=maintype,
                    subtype=subtype,
                    filename=fn
                )
            else:
                attch_payload = MIMEText(base64_encode(data) + "\r\n", "base64")
                attch_payload.set_charset("utf-8")
                attch_payload.add_header("Content-Disposition", "attachment", filename=fn)
                attch_payload.add_header("Content-Transfer-Encoding", "base64")
                self.message_obj._payload[0].attach(attch_payload)

        signed_payload = self.pgp_config.lock(CustomMIMEWrapper(self.message_obj))
        conn = self.test_conn(username, password, host, port, verify_conn)
        conn.sendmail(self.message_obj["From"], self.message_obj["To"], signed_payload)

    def get_summary(self):
        return {
            "message_type": self.pgp_config.pgp_action or "Cleartext",
            "sent_to": self.message_obj["To"],
            "num_attachments": len(self._attachments.attachments),
            "attach_filenames": ",".join([a.filename for a in self._attachments.attachments]),
            "attach_hashes_md5": ",".join([a.hash_md5 for a in self._attachments.attachments]),
            "attach_hashes_sha1": ",".join([a.hash_sha256 for a in self._attachments.attachments]),
            "attach_hashes_sha256": ",".join([a.hash_sha256 for a in self._attachments.attachments]),
            "subject": self.message_obj["Subject"]
        }



import re
import email
import hashlib
from email_master.util import EmailParser, EmailUtil, EmailAttachment, EmailAttachmentList
from email_master.compat import base64_decode, base64_decode_to_bytes, to_unicode
import six


if six.PY2:
    from HTMLParser import HTMLParser
    from types import StringTypes
elif six.PY3:
    from html.parser import HTMLParser
    StringTypes = str
    unicode = to_unicode


class EMLParser(EmailParser):
    def __init__(self, email_data, filename="unknown.eml", ignore_errors=False, exclude_attachment_extensions=None):
        super(EMLParser, self).__init__(email_data,
                                        filename=filename,
                                        ignore_errors=ignore_errors,
                                        exclude_attachment_extensions=exclude_attachment_extensions)
        if six.PY3:
            self.msg = email.message_from_bytes(base64_decode_to_bytes(email_data))
        else:
            self.msg = email.message_from_string(base64_decode(email_data))

        self.seen_attachment_hashes = []  # Class-global list to keep track of which attachments have been processed

    def get_cc(self):
        return self.msg["cc"]

    def get_bcc(self):
        return self.msg["bcc"]

    def get_raw_content(self):
        if six.PY2:
            return self.msg.as_string()
        else:
            charsets = self.msg.get_charsets()
            for charset in charsets:
                to_try = EmailUtil.validate_charset(charset)
                try:
                    return bytes(self.msg).decode(to_try, errors="replace")
                except:
                    continue

            # Can't figure out the encoding, try utf-8 with ignore, if that fails, it will raise exception
            return bytes(self.msg).decode("utf-8", errors="ignore")

    def get_raw_headers(self):
        return u"\n".join([u"{}: {}".format(unicode(h[0], "utf-8", "ignore"), unicode(h[1], "utf-8", "ignore")) for h in self.msg.items() or []])

    def get_sender(self):
        return EmailUtil.try_decode(self.msg['From']) or u""

    def get_reply_to(self):
        return EmailUtil.try_decode(self.msg["Reply-To"]) or u""

    def get_plaintext_body(self):
        if self.msg.is_multipart():
            for part in self.msg.walk():
                ctype = part.get_content_type()
                cdispo = str(part.get('Content-Disposition'))
                if ctype == 'text/plain' and 'attachment' not in cdispo:
                    return self._decode_body(part)
        else:
            if self.msg.get_content_type() == "text/plain":
                return self._decode_body(self.msg)

        return u""

    def _decode_body(self, part):
        body = None
        for ch in part.get_charsets():
            ch = EmailUtil.validate_charset(ch)
            if ch == "utf-8":
                continue  # Wait to default to utf-8

            body = part.get_payload(decode=True).decode(ch, errors="replace")

        if not body:
            body = part.get_payload(decode=True).decode("utf-8", errors="replace")

        return body

    def get_html_body(self, decode_html=True):
        if self.msg.is_multipart():
            for part in self.msg.walk():
                ctype = part.get_content_type()
                cdispo = str(part.get('Content-Disposition'))
                if ctype == 'text/html' and 'attachment' not in cdispo:
                    # Try decoding HTML Entities
                    body = self._decode_body(part)
                    if decode_html:
                        return HTMLParser().unescape(body) or u""
                    else:
                        return body or u""
        else:
            if self.msg.get_content_type() == "text/html":
                return self._decode_body(self.msg)

        return u""

    def get_rtf_body(self):
        return u""  # Pretty sure EML files can't/shouldn't have rtf bodies

    def get_subject(self):
        return EmailUtil.try_decode(self.msg['Subject']) if self.msg['Subject'] else u"(No Subject)"

    def get_type(self):
        return u"EML"

    def get_date(self):
        return self.msg['Date'] or ""

    def get_recipients(self):
        return EmailUtil.try_decode(self.msg['To']) or ""

    def get_id(self):
        return self.msg['id'] or self.msg['Message-ID'] or ""

    def get_headers(self):
        parser = email.parser.HeaderParser()
        headers = parser.parsestr(self.get_raw_content()).items()
        return headers if headers else []

    def get_attachments(self):
        """
        Handle the attachments and return strings of 'attachments_md5', 'attachments_sha1' and 'attach_info', the hashed
        data of the attachments, where attach_info is the Content-Type and Content-Transfer-Encoding concatenated
        :return: dictionary with csv values
        """
        raw_attachments = self._organize_attachments(self.msg)
        attachments = EmailAttachmentList()

        for raw_attachment in raw_attachments:
            if isinstance(raw_attachment[1], StringTypes) and six.PY3:
                attach_data = raw_attachment[1].encode()
            else:
                attach_data = raw_attachment[1]

            filename = self._get_attachment_filename(raw_attachment[0], hashlib.md5(attach_data).hexdigest())

            email_attachment = EmailAttachment(
                u" ".join([u"{}: {}".format(unicode(k, "utf-8", "ignore"), unicode(v, "utf-8", "ignore")) for k, v in raw_attachment[0].items()]),
                filename, raw_attachment[1])

            # Make sure that the attachments aren't the same as the html_body or plaintext_body
            email_attachment.raw_data = email_attachment.raw_data.decode("utf-8", errors="replace")
            bodies = (self.get_html_body(decode_html=False), self.get_plaintext_body(), self.get_rtf_body())
            if email_attachment.raw_data not in bodies:
                attachments.add_attachment(email_attachment)

        return attachments

    @staticmethod
    def _get_attachment_filename(attachment_headers, fallback_name):
        """
        Attempt to get the filename from the content-type information, otherwise just use the fallback name
        :param attachment_headers: List of headers for the attachment
        :param fallback_name: Name to use if we can't find a filename, will be prefixed with unknown-<name>
        :return: filename to use for the attachment
        """
        attachment_headers = {k.lower(): v for k, v in attachment_headers.items()}
        filename = "unknown-{}".format(fallback_name)
        if "content-location" in attachment_headers:
            filename = attachment_headers["content-location"]
        elif 'subject' in attachment_headers:
            filename = "{}.eml".format(attachment_headers['subject'])
        elif "content-disposition" in attachment_headers:
            content = attachment_headers["content-disposition"].split(";")
            for disp in content:
                disp = disp.strip()
                if disp.startswith("filename"):
                    match = re.findall("filename=\"?([^\"]*)\"?", disp)
                    if match:
                        filename = match[0]
                    break
        elif "content-type" in attachment_headers:
            properties = attachment_headers["content-type"].split(";")
            for prop in properties:
                prop = prop.strip()  # Strip \t from prop
                if prop.startswith("name") or prop.startswith("filename"):
                    # Split 'name="asdf.png"' into 'asdf.png'
                    fn = re.split("name=\"?", prop)
                    if len(fn) > 1:  # If split failed we can't determine filename
                        fn = fn[1]
                        if fn.endswith("\"") or fn.endswith("\'"):
                            filename = fn[:-1]  # Cut off trailing quote
                    break

        return EmailUtil.try_decode(filename)  # Try and decode filenames that are encoded with mime encoded-word syntax

    @staticmethod
    def _is_attachment(headers):
        """
        Check if a given payload is an attachment (or it's email body)
        Valid from https://www.w3.org/Protocols/rfc1341/5_Content-Transfer-Encoding.html
        :param headers:
        :return:
        """
        if "content-disposition" in headers:
            if headers["content-disposition"].startswith("attachment"):
                return True
        if "content-transfer-encoding" in headers:
            trans_enc = headers["content-transfer-encoding"].lower()
            if trans_enc in ("quoted-printable", "7bit", "8bit"):
                return False
            else:
                return True

    def _organize_attachments(self, eml_obj):
        """
        Organize and filter the attachment data
        :param eml_obj: email object to organize
        :return: List like [[headers, attachment], ...]
        """
        data = self._extract_attachments(eml_obj)  # Recursively extract attachments from EML
        data = filter(lambda x: bool(x), data)  # Filter out the None attachments
        return data

    def _parse_attachment(self, headers, attachment):
        """
        Take in a list of headers and a Message object, and pre-parse the attachment
        Returns None if it isn't an attachment
        :param headers: [(header1, value1), ...]
        :param attachment: Message object to parse
        :return: [{headers}, <attachment raw data>
        """
        lower_headers = {}
        for k, v in headers:
            lower_headers[k.lower()] = v
        if self._is_attachment(lower_headers):
            newdata = self._decode_payload(attachment)
            return [lower_headers, newdata]
        else:
            return None

    def _extract_attachments(self, eml_obj):
        """
        Recursively parse attachments from an email object, keeping track of parsed emails using
        self.seen_attachment_hashes

        :param eml_obj: Message object
        :return: [ [{headers}, <attachment raw data>], ...]
        """
        if hash(eml_obj) in self.seen_attachment_hashes:
            return []
        else:
            data = []
            payload = eml_obj.get_payload()
            # Very specific case of the attached email being an EML that has been parsed into an attachment
            if isinstance(payload, list) and len(payload) == 1 and isinstance(payload[0], email.message.Message)\
                    and isinstance(payload[0].get_payload(), StringTypes):
                attachment = payload[0]
                if not hash(attachment) in self.seen_attachment_hashes:
                    parent_headers = eml_obj.items()
                    child_headers = attachment.items()
                    combined_headers = []  # Combine headers to give us the most information about this possible attachment
                    combined_headers.extend(parent_headers)
                    combined_headers.extend(child_headers)
                    data.append(self._parse_attachment(combined_headers, attachment))
                    self.seen_attachment_hashes.append(hash(attachment))
            elif isinstance(payload, list):
                for attachment in payload:
                    if isinstance(attachment, email.message.Message):
                        # First condition checks whether the attachment is another email
                        # by checking original and new msg ids
                        msg_id = attachment.get('id') or attachment.get('Message-id')
                        if self.get_id() is not msg_id and msg_id is not None:
                            data.append([dict(attachment.items()), str(eml_obj)])
                        elif not hash(attachment) in self.seen_attachment_hashes:
                            data.extend(self._extract_attachments(attachment))
                            self.seen_attachment_hashes.append(hash(attachment))
                        else:
                            continue  # Skip, we've processed this before
                    else:
                        # This shouldn't be an attachment, but in case it is, add it and filter after
                        data.append(self._parse_attachment(eml_obj.items(), attachment))
            elif isinstance(payload, StringTypes):
                data.append(self._parse_attachment(eml_obj.items(), eml_obj))
            else:
                pass
                # Shouldn't ever be here!
        return data

    @staticmethod
    def _decode_payload(attachment):
        """
        Take a given email payload and get the raw unencoded data
        :param attachment: email attachment part
        :return:
        """
        data = attachment.get_payload()
        if data and isinstance(data[0], email.message.Message):
            data = data[0].get_payload()  # Email-ception

        if "content-transfer-encoding" in attachment:
            if attachment["Content-Transfer-Encoding"].lower() == "base64":
                try:
                    data = base64_decode_to_bytes(data)
                except TypeError:
                    data = base64_decode_to_bytes(data, altchars="-_")
        return data



from email_master.eml import EMLParser
from email_master.msg import MSGParser
from email_master.util import EmailParser
from extract_msg.exceptions import InvalidFileFormat
from email_master.exceptions import InvalidDataException

import six


if six.PY3:
    # TODO MSG library throws the first three errors if it isn't an MSG in python3, refactor check or update lib?
    ExceptionList = (IOError, OSError, AttributeError, InvalidDataException, InvalidFileFormat)
else:
    ExceptionList = (InvalidDataException, InvalidFileFormat)


class EmailMasterParser(EmailParser):
    def _get_parser(self, email_data, file_parsers=None, **kwargs):
        if not file_parsers:
            raise Exception("Need a list of classtypes to attempt parsing!")

        errors = []
        # Try the different parsers for email attachments)
        for file_parser in file_parsers:
            try:
                return file_parser(email_data, **kwargs)
            except ExceptionList as e:
                errors.append("{}: {}".format(file_parser.__class__.__name__, str(e)))
                continue
        raise Exception("Unable to parse message, errors: {}".format(str(errors)))

    def __init__(self, email_data, filename="unknown", parser_classes=None, **kwargs):
        """

        :param email_data: Email data suitable for use in the constructor of an EmailParser class instance
        :param kwargs: ignore_errors = False, exclude_attachment_extensions = None
        :param parser_classes: List of ClassTypes to try parsing with, defaults to [MSGParser, EMLParser]
        """

        if not parser_classes:
            parser_classes = [MSGParser, EMLParser]

        self.parser = self._get_parser(email_data, filename=filename, file_parsers=parser_classes, **kwargs)
        super(EmailMasterParser, self).__init__(email_data, filename=filename, **kwargs)

    def get_cc(self):
        return self.parser.get_cc()

    def get_bcc(self):
        return self.parser.get_bcc()

    def get_raw_content(self):
        return self.parser.get_raw_content()

    def get_raw_headers(self):
        return self.parser.get_raw_headers()

    def get_reply_to(self):
        return self.parser.get_reply_to()

    def get_sender(self):
        return self.parser.get_sender()

    def get_plaintext_body(self):
        return self.parser.get_plaintext_body()

    def get_html_body(self, decode_html=True):
        return self.parser.get_html_body()

    def get_rtf_body(self):
        return self.parser.get_rtf_body()

    def get_subject(self):
        return self.parser.get_subject()

    def get_type(self):
        return self.parser.get_type()

    def get_recipients(self):
        return self.parser.get_recipients()

    def get_headers(self):
        return self.parser.get_headers()

    def get_attachments(self):
        return self.parser.get_attachments()

    def get_id(self):

        mid = self.parser.get_id()
        mid = mid.strip('\t\n\a')
        return mid

    def get_date(self):
        return self.parser.get_date()




from email_master.pgp import PGPConfig
from email_master.attachments import EmailAttachmentList, EmailAttachment
from email_master.compat import base64_decode_to_bytes


class EmailMasterMessage(object):
    def __init__(self, message_obj, pgp_config=None):
        self.message_obj = message_obj
        if pgp_config is None:
            self.pgp_config = PGPConfig()
        else:
            self.pgp_config = pgp_config

        self._attachments = EmailAttachmentList()

    def test_conn(self, username, password, host, port, verify_conn=True):
        raise NotImplementedError

    def send(self, username, password, host, port, verify_conn=True):
        raise NotImplementedError

    def add_attachments(self, attachs):
        if isinstance(attachs, EmailAttachmentList):
            for att in attachs.attachments:
                self._attachments.add_attachment(att)
        elif isinstance(attachs, list):
            for att in attachs:
                self.add_attachment(att)
        else:
            raise ValueError("Invalid input! Must be type list or EmailAttachmentList, for singular attachments, use add_attachment!")

    def add_attachment(self, attachment):
        if isinstance(attachment, EmailAttachment):
            self._attachments.add_attachment(attachment)
        else:
            if not isinstance(attachment, dict):
                raise ValueError("Attachment must be dict or EmailAttachment!")

            self._attachments.add_attachment(
                EmailAttachment({}, attachment["filename"], base64_decode_to_bytes(attachment["base64"]))
            )

    def get_summary(self):
        raise NotImplementedError



from setuptools import setup


with open('./README.rst') as f:
    long_description = f.read()

setup(
    name='email-master',
    packages=['email_master', 'email_master.eml', 'email_master.msg'],
    version='0.3.16',
    description='Master Email Parsing Package',
    author='Swimlane',
    author_email="info@swimlane.com",
    long_description=long_description,
    install_requires=[
        "pendulum==1.2.5",
        "compressed-rtf==1.0.5",
        "extract_msg==0.23.2",
        "six==1.13.0",
        "PGPy==0.5.2"
    ],
    keywords=['utilities', 'email', 'parsing', 'eml', 'msg'],
    classifiers=[],
)
# 12, 13, 14, 15, 16